#!/usr/bin/env python

import copy

from mapel.core.objects.Family import Family
from mapel.roommates.objects.Roommates import Roommates
from mapel.core.utils import *

import mapel.roommates.cultures.mallows as mallows


class RoommatesFamily(Family):

    def __init__(self,
                 culture_id: str = None,
                 family_id='none',
                 params: dict = None,
                 size: int = 1,
                 label: str = "none",
                 color: str = "black",
                 alpha: float = 1.,
                 ms: int = 20,
                 show=True,
                 marker='o',
                 starting_from: int = 0,
                 path: dict = None,
                 single: bool = False,
                 num_agents: int = None):

        super().__init__(culture_id=culture_id,
                         family_id=family_id,
                         params=params,
                         size=size,
                         label=label,
                         color=color,
                         alpha=alpha,
                         ms=ms,
                         show=show,
                         marker=marker,
                         starting_from=starting_from,
                         path=path,
                         single=single)

        self.num_agents = num_agents

    def _get_params_for_paths(self, j, extremes=False):
        path = self.path

        variable = path['variable']

        if 'extremes' in path:
            extremes = path['extremes']

        params = {'variable': variable}
        if extremes:
            params[variable] = j / (self.size - 1)
        elif not extremes:
            params[variable] = (j + 1) / (self.size + 1)

        if 'scale' in path:
            params[variable] *= path['scale']

        if 'start' in path:
            params[variable] += path['start']
        else:
            path['start'] = 0.

        if 'step' in path:
            params[variable] = path['start'] + j * path['step']

        return params, variable

    def prepare_family(self, experiment_id=None, is_exported=None):

        instances = {}

        _keys = []
        for j in range(self.size):

            params = copy.deepcopy(self.params)

            path = self.path
            if path is not None and 'variable' in path:
                new_params, variable = self._get_params_for_paths(j)
                params = {**params, **new_params}

            if params is not None and 'normphi' in params:
                params['phi'] = mallows.phi_from_normphi(self.num_agents,
                                                        relphi=params['normphi'])

            instance_id = get_instance_id(self.single, self.family_id, j)

            instance = Roommates(experiment_id, instance_id, is_imported=False,
                                 culture_id=self.culture_id, num_agents=self.num_agents)

            instance.prepare_instance(is_exported=is_exported, params=params)

            instances[instance_id] = instance

            _keys.append(instance_id)

        self.instance_ids = _keys

        return instances



import numpy as np
from mapel.roommates.cultures._utils import convert


def generate_roommates_urn_votes(num_agents: int = None,
                                 alpha: int = 0.1,
                                 **kwargs):

    votes = np.zeros([num_agents, num_agents], dtype=int)
    urn_size = 1.
    for j in range(num_agents):
        rho = np.random.uniform(0, urn_size)
        if rho <= 1.:
            votes[j] = np.random.permutation(num_agents)
        else:
            votes[j] = votes[np.random.randint(0, j)]
        urn_size += alpha

    return convert(votes)

# # # # # # # # # # # # # # # #
# LAST CLEANUP ON:  9.06.2023 #
# # # # # # # # # # # # # # # #




from mapel.core.glossary import *
from mapel.core.utils import *
from collections import Counter


def export_instance_to_a_file(experiment):
    """ Store votes in a file """

    path_to_folder = os.path.join(os.getcwd(), "election", experiment.experiment_id, "instances")
    make_folder_if_do_not_exist(path_to_folder)
    path_to_file = os.path.join(path_to_folder, f'{experiment.instance_id}.ri')

    with open(path_to_file, 'w') as file_:

        if experiment.culture_id in NICE_NAME:
            file_.write("# " + NICE_NAME[experiment.culture_id] + " " + str(experiment.params) + "\n")
        else:
            file_.write("# " + experiment.culture_id + " " + str(experiment.params) + "\n")

        file_.write(str(experiment.num_agents) + "\n")

        for i in range(experiment.num_agents):
            file_.write(str(i) + ', a' + str(i) + "\n")

        c = Counter(map(tuple, experiment.votes))
        counted_votes = [[count, list(row)] for row, count in c.items()]

        file_.write(str(experiment.num_agents) + ', ' + str(experiment.num_agents) + ', ' +
                    str(len(counted_votes)) + "\n")

        for i in range(len(counted_votes)):
            file_.write(str(counted_votes[i][0]) + ', ')
            for j in range(len(counted_votes[i][1])):
                file_.write(str(int(counted_votes[i][1][j])))
                if j < len(counted_votes[i][1]) - 1:
                    file_.write(", ")
            file_.write("\n")




#!/usr/bin/env python
import copy
import csv
from time import time
import logging
import os
from typing import Callable

import numpy as np

from mapel.core.inner_distances import map_str_to_func
from mapel.core.objects.Experiment import Experiment
from mapel.roommates.distances import main_distances as mrd
from mapel.roommates.objects.Roommates import Roommates

registered_roommates_distances = {
    'mutual_attraction': mrd.compute_retrospective_distance,

    'positionwise': mrd.compute_positionwise_distance,  # unsupported distance
    'pos_swap': mrd.compute_pos_swap_distance,  # unsupported distance
    'swap_bf': mrd.compute_swap_bf_distance,  # unsupported distance
    'pairwise': mrd.compute_pairwise_distance,  # unsupported distance
}


def get_distance(election_1: Roommates, election_2: Roommates,
                 distance_id: str = None) -> float or (float, list):
    """ Return: distance between ordinal elections, (if applicable) optimal matching """

    inner_distance, main_distance = extract_distance_id(distance_id)

    if main_distance in registered_roommates_distances:
        return registered_roommates_distances.get(main_distance)(election_1,
                                                                 election_2,
                                                                 inner_distance)
    else:
        logging.warning('No such metric!')


def extract_distance_id(distance_id: str) -> (Callable, str):
    if '-' in distance_id:
        inner_distance, main_distance = distance_id.split('-')
        inner_distance = map_str_to_func(inner_distance)
    else:
        main_distance = distance_id
        inner_distance = None
    return inner_distance, main_distance


def run_single_thread(experiment: Experiment,
                      thread_ids: list,
                      distances: dict,
                      times: dict,
                      matchings: dict,
                      t) -> None:
    """ Single thread for computing distances """
    for election_id_1, election_id_2 in thread_ids:
        start_time = time()

        distance = get_distance(copy.deepcopy(experiment.instances[election_id_1]),
                                copy.deepcopy(experiment.instances[election_id_2]),
                                distance_id=copy.deepcopy(experiment.distance_id),
                                )
        if type(distance) is tuple:
            distance, matching = distance
            matching = np.array(matching)
            matchings[election_id_1][election_id_2] = matching
            matchings[election_id_2][election_id_1] = np.argsort(matching)
        distances[election_id_1][election_id_2] = distance
        distances[election_id_2][election_id_1] = distances[election_id_1][election_id_2]
        times[election_id_1][election_id_2] = time() - start_time
        times[election_id_2][election_id_1] = times[election_id_1][election_id_2]

    if experiment.is_exported:

        file_name = f'{experiment.distance_id}_p{t}.csv'
        path = os.path.join(os.getcwd(), "election", experiment.experiment_id, "distances",
                            file_name)

        with open(path, 'w', newline='') as csv_file:
            writer = csv.writer(csv_file, delimiter=';')
            writer.writerow(
                ["instance_id_1", "instance_id_2", "distance", "time"])

            for election_id_1, election_id_2 in thread_ids:
                distance = float(distances[election_id_1][election_id_2])
                time_ = float(times[election_id_1][election_id_2])
                writer.writerow([election_id_1, election_id_2, distance, time_])

# # # # # # # # # # # # # # # #
# LAST CLEANUP ON: 13.10.2021 #
# # # # # # # # # # # # # # # #



#!/usr/bin/env python
import ast
import copy
import csv
import itertools
from multiprocessing import Process
from time import sleep
import time

from mapel.core.objects.Experiment import Experiment
from mapel.roommates.objects.RoommatesFamily import RoommatesFamily
from mapel.roommates.objects.Roommates import Roommates
import mapel.roommates.cultures_ as models_main
import mapel.roommates.distances_ as metr
import mapel.roommates.features.basic_features as basic
import mapel.roommates.features_ as features
from mapel.core.utils import *
from mapel.core.glossary import *
from mapel.core.persistence.experiment_imports import get_values_from_csv_file

try:
    from sklearn.manifold import MDS
    from sklearn.manifold import TSNE
    from sklearn.manifold import SpectralEmbedding
    from sklearn.manifold import LocallyLinearEmbedding
    from sklearn.manifold import Isomap
except ImportError as error:
    MDS = None
    TSNE = None
    SpectralEmbedding = None
    LocallyLinearEmbedding = None
    Isomap = None
    print(error)


class RoommatesExperiment(Experiment):
    """Abstract set of elections."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.default_num_agents = 10
        self.matchings = {}
        try:
            self.import_matchings()
        except:
            pass

    def import_matchings(self):
        matchings = {}

        path = os.path.join(os.getcwd(), 'election', self.experiment_id, 'features',
                            'stable_sr.csv')
        with open(path, 'r', newline='') as csv_file:
            reader = csv.DictReader(csv_file, delimiter=';')

            for row in reader:
                election_id = row['instance_id']
                value = row['matching']
                if value == '':
                    matchings[election_id] = None
                else:
                    matchings[election_id] = value

        self.matchings = matchings

    def add_instances_to_experiment(self):

        instances = {}

        for family_id in self.families:
            single = self.families[family_id].single

            ids = []
            for j in range(self.families[family_id].size):
                instance_id = get_instance_id(single, family_id, j)
                instance = Roommates(self.experiment_id, instance_id)
                instances[instance_id] = instance
                ids.append(str(instance_id))

            self.families[family_id].instance_ids = ids

        return instances

    def add_instance(self, culture_id="none", params=None, label=None,
                     color="black", alpha=1., show=True, marker='x', starting_from=0, size=1,
                     num_agents=None, instance_id=None):

        if num_agents is None:
            num_agents = self.default_num_agents

        return self.add_family(culture_id=culture_id, params=params, size=size, label=label,
                               color=color, alpha=alpha, show=show, marker=marker,
                               starting_from=starting_from, family_id=instance_id,
                               num_agents=num_agents, single_instance=True)

    def add_family(self, culture_id: str = "none", params: dict = None, size: int = 1,
                   label: str = None, color: str = "black", alpha: float = 1.,
                   show: bool = True, marker: str = 'o', starting_from: int = 0,
                   family_id: str = None, single_instance: bool = False,
                   num_agents: int = None, path: dict = None):

        if num_agents is None:
            num_agents = self.default_num_agents

        if self.families is None:
            self.families = {}

        if label is None:
            label = family_id

        self.families[family_id] = RoommatesFamily(culture_id=culture_id, family_id=family_id,
                                                   params=params, label=label, color=color,
                                                   alpha=alpha, single=single_instance,
                                                   show=show, size=size, marker=marker,
                                                   starting_from=starting_from,
                                                   num_agents=num_agents, path=path)

        self.num_families = len(self.families)
        self.num_instances = sum([self.families[family_id].size for family_id in self.families])

        models_main.prepare_instances(experiment=self,
                                    culture_id=self.families[family_id].culture_id,
                                    family_id=family_id,
                                    params=copy.deepcopy(self.families[family_id].params))

    def compute_distances(self,
                          distance_id: str = 'emd-positionwise',
                          num_threads: int = 1,
                          self_distances: bool = False) -> None:

        self.distance_id = distance_id

        if '-pairwise' in distance_id:
            for instance in self.instances.values():
                instance.votes_to_pairwise_matrix()

        matchings = {instance_id: {} for instance_id in self.instances}
        distances = {instance_id: {} for instance_id in self.instances}
        times = {instance_id: {} for instance_id in self.instances}

        ids = []
        for i,instance_1 in enumerate(self.instances):
            for j, instance_2 in enumerate(self.instances):
                if i == j:
                    if self_distances:
                        ids.append((instance_1, instance_2))
                elif i < j:
                    ids.append((instance_1, instance_2))

        num_distances = len(ids)
        processes =[]

        for t in range(num_threads):
            print(f'Starting thread: {t}')
            sleep(0.1)
            start = int(t * num_distances / num_threads)
            stop = int((t + 1) * num_distances / num_threads)
            thread_ids = ids[start:stop]

            process = Process(target=metr.run_single_thread, args=(self,
                                                                   thread_ids,
                                                                     distances,
                                                                   times,
                                                                   matchings,
                                                                     t ))

            process.start()
            processes.append(process)

        for process in processes:
            process.join()

        distances = {instance_id: {} for instance_id in self.instances}
        times = {instance_id: {} for instance_id in self.instances}
        for t in range(num_threads):

            file_name = f'{distance_id}_p{t}.csv'
            path = os.path.join(os.getcwd(), "election", self.experiment_id, "distances",
                                file_name)

            with open(path, 'r', newline='') as csv_file:
                reader = csv.DictReader(csv_file, delimiter=';')

                for row in reader:
                    distances[row['instance_id_1']][row['instance_id_2']] = float(row['distance'])
                    times[row['instance_id_1']][row['instance_id_2']] = float(row['time'])

        if self.is_exported:

            path_to_folder = os.path.join(os.getcwd(), "election", self.experiment_id,
                                          "distances")
            make_folder_if_do_not_exist(path_to_folder)
            path_to_file = os.path.join(path_to_folder, f'{distance_id}.csv')

            with open(path_to_file, 'w', newline='') as csv_file:
                writer = csv.writer(csv_file, delimiter=';')
                writer.writerow(
                    ["instance_id_1", "instance_id_2", "distance", "time"])

                for election_1, election_2 in itertools.combinations(self.instances, 2):
                    distance = str(distances[election_1][election_2])
                    time = str(times[election_1][election_2])
                    writer.writerow([election_1, election_2, distance, time])

        self.distances = distances
        self.times = times
        self.matchings = matchings

        for instance_id_1 in self.distances:
            for instance_id_2 in self.distances[instance_id_1]:
                self.distances[instance_id_2][instance_id_1] = \
                    self.distances[instance_id_1][instance_id_2]
                self.times[instance_id_2][instance_id_1] = \
                    self.times[instance_id_1][instance_id_2]

    def import_controllers(self):
        """ Import controllers from a file """

        families = {}

        path = os.path.join(os.getcwd(), 'experiments', self.experiment_id, 'map.csv')
        file_ = open(path, 'r')

        header = [h.strip() for h in file_.readline().split(';')]
        reader = csv.DictReader(file_, fieldnames=header, delimiter=';')

        starting_from = 0
        for row in reader:

            culture_id = None
            color = None
            label = None
            params = None
            alpha = None
            size = None
            marker = None
            num_agents = None
            family_id = None
            show = True

            if 'culture_id' in row.keys():
                culture_id = str(row['culture_id']).strip()

            if 'color' in row.keys():
                color = str(row['color']).strip()

            if 'label' in row.keys():
                label = str(row['label'])

            if 'family_id' in row.keys():
                family_id = str(row['family_id'])

            if 'params' in row.keys():
                params = ast.literal_eval(str(row['params']))

            if 'alpha' in row.keys():
                alpha = float(row['alpha'])

            if 'size' in row.keys():
                size = int(row['size'])

            if 'marker' in row.keys():
                marker = str(row['marker']).strip()

            if 'num_agents' in row.keys():
                num_agents = int(row['num_agents'])

            if 'path' in row.keys():
                path = ast.literal_eval(str(row['path']))

            if 'show' in row.keys():
                show = row['show'].strip() == 't'

            single_instance = size == 1

            families[family_id] = RoommatesFamily(culture_id=culture_id,
                                                  family_id=family_id,
                                                  params=params, label=label,
                                                  color=color, alpha=alpha, show=show,
                                                  size=size, marker=marker,
                                                  starting_from=starting_from,
                                                  num_agents=num_agents, path=path,
                                                  single=single_instance)
            starting_from += size

        self.num_families = len(families)
        self.num_instances = sum([families[family_id].size for family_id in families])
        self.main_order = [i for i in range(self.num_instances)]

        file_.close()
        return families

    def prepare_instances(self):

        if self.instances is None:
            self.instances = {}

        for family_id in self.families:

            new_instances = self.families[family_id].prepare_family(
                is_exported=self.is_exported,
                experiment_id=self.experiment_id)

            for instance_id in new_instances:
                self.instances[instance_id] = new_instances[instance_id]

    def compute_stable_sr(self):
        for instance_id in self.instances:
            print(instance_id)
            if instance_id in ['roommates_test']:
                self.matchings[instance_id] = 'None'
            else:
                usable_matching = basic.compute_stable_SR(self.instances[instance_id].votes)
                self.matchings[instance_id] = usable_matching

        if self.is_exported:

            path_to_folder = os.path.join(os.getcwd(), "election", self.experiment_id,
                                          "features")
            make_folder_if_do_not_exist(path_to_folder)
            path_to_file = os.path.join(path_to_folder, f'stable_sr.csv')

            with open(path_to_file, 'w', newline='') as csv_file:
                writer = csv.writer(csv_file, delimiter=';')
                writer.writerow(
                    ["instance_id", "matching"])

                for instance_id in self.instances:
                    usable_matching = self.matchings[instance_id]
                    writer.writerow([instance_id, usable_matching])

    def compute_feature(self,
                        feature_id: str = None,
                        feature_params=None) -> dict:

        if feature_params is None:
            feature_params = {}

        feature_dict = {'value': {}, 'time': {}, 'std': {}}

        features_with_std = {'avg_num_of_bps_for_rand_matching'}

        num_iterations = 1
        if 'num_iterations' in feature_params:
            num_iterations = feature_params['num_iterations']

        if feature_id in MAIN_GLOBAL_FEATUERS:

            feature = features.get_global_feature(feature_id)

            values = feature(self, election_ids=list(self.instances))

            for instance_id in self.instances:
                feature_dict['value'][instance_id] = values[instance_id]
                feature_dict['time'][instance_id] = 0

        elif feature_id == 'distortion_from_all':
            feature = features.get_global_feature(feature_id)
            for instance_id in self.instances:
                values = feature(self, self.instances[instance_id])
                print(instance_id, values)
                feature_dict['value'][instance_id] = values
                feature_dict['time'][instance_id] = 0
        else:

            if feature_id == 'summed_rank_difference':
                minimal = get_values_from_csv_file(self, feature_id='summed_rank_minimal_matching')
                maximal = get_values_from_csv_file(self, feature_id='summed_rank_maximal_matching')

                for instance_id in self.instances:
                    if minimal[instance_id] is None:
                        value = 'None'
                    else:
                        value = abs(maximal[instance_id] - minimal[instance_id])
                    feature_dict['value'][instance_id] = value
                    feature_dict['time'][instance_id] = 0

            else:
                for instance_id in self.instances:
                    feature = features.get_local_feature(feature_id)
                    instance = self.instances[instance_id]

                    start = time.time()

                    for _ in range(num_iterations):

                        if feature_id in ['summed_rank_minimal_matching',
                                          'summed_rank_maximal_matching',
                                          'minimal_rank_maximizing_matching',
                                          ] \
                                and (self.matchings[instance_id] is None or self.matchings[instance_id] == 'None'):
                            value = 'None'
                        else:
                            value = feature(instance)

                        print(value)

                    total_time = time.time() - start
                    total_time /= num_iterations

                    if feature_id in features_with_std:
                        feature_dict['value'][instance_id] = value[0]
                        feature_dict['time'][instance_id] = total_time
                        feature_dict['std'][instance_id] = value[1]
                    else:
                        feature_dict['value'][instance_id] = value
                        feature_dict['time'][instance_id] = total_time

        if self.is_exported:

            path_to_folder = os.path.join(os.getcwd(), "election", self.experiment_id,
                                          "features")
            make_folder_if_do_not_exist(path_to_folder)
            path_to_file = os.path.join(path_to_folder, f'{feature_id}.csv')

            with open(path_to_file, 'w', newline='') as csv_file:
                writer = csv.writer(csv_file, delimiter=';')

                if feature_id in features_with_std:
                    writer.writerow(["election_id", "value", 'time', 'std'])
                    for key in feature_dict['value']:
                        writer.writerow([key, feature_dict['value'][key],
                                         round(feature_dict['time'][key],3),
                                        round(feature_dict['std'][key],3)])
                else:
                    writer.writerow(["election_id", "value", 'time'])
                    for key in feature_dict['value']:
                        writer.writerow([key, feature_dict['value'][key],
                                         round(feature_dict['time'][key],3)])

        self.features[feature_id] = feature_dict
        return feature_dict

    def create_structure(self) -> None:

        dirs = ["experiments", "images", "trash"]
        for dir in dirs:
            if not os.path.isdir(dir):
                os.mkdir(os.path.join(os.getcwd(), dir))

        try:
            os.mkdir(os.path.join(os.getcwd(), "election", self.experiment_id))
            os.mkdir(os.path.join(os.getcwd(), "election", self.experiment_id, "distances"))
            os.mkdir(os.path.join(os.getcwd(), "election", self.experiment_id, "features"))
            os.mkdir(os.path.join(os.getcwd(), "election", self.experiment_id, "coordinates"))
            os.mkdir(os.path.join(os.getcwd(), "election", self.experiment_id, "instances"))
            os.mkdir(os.path.join(os.getcwd(), "election", self.experiment_id, "matrices"))

            # PREPARE MAP.CSV FILE
            path = os.path.join(os.getcwd(), "election", self.experiment_id, "map.csv")

            with open(path, 'w') as file_csv:
                file_csv.write(
                    "size;num_agents;culture_id;params;color;alpha;family_id;label;marker;show\n")
                file_csv.write("10;20;roommates_ic;{};black;1;IC;IC;o;t\n")
        except:
            pass







import os
import ast

from mapel.core.glossary import NICE_NAME


def import_real_instance(self, shift=False):
    """ Import real ordinal election form .soc file """

    file_name = f'{self.instance_id}.ri'
    path = os.path.join(os.getcwd(), "election", self.experiment_id, "instances", file_name)
    print(path)
    with open(path, 'r') as my_file:
        params = 0
        first_line = my_file.readline()

        if first_line[0] != '#':
            culture_id = 'empty'
            num_agents = int(first_line)
        else:
            first_line = first_line.strip().split()
            culture_id = first_line[1]

            if len(first_line) <= 2:
                params = {}
            else:
                params = ast.literal_eval(" ".join(first_line[2:]))

            num_agents = int(my_file.readline())

        num_candidates = num_agents - 1

        for _ in range(num_agents):
            my_file.readline()

        line = my_file.readline().rstrip("\n").split(',')
        num_voters = int(line[0])
        num_options = int(line[2])
        votes = [[0 for _ in range(num_candidates)] for _ in range(num_voters)]

        it = 0
        for j in range(num_options):

            line = my_file.readline().rstrip("\n").split(',')
            quantity = int(line[0])

            for k in range(quantity):
                print(k, votes)
                for el in range(num_candidates):
                    votes[it][el] = int(line[el + 1])
                it += 1

        if shift:
            for i in range(num_voters):
                for j in range(num_candidates):
                    votes[i][j] -= 1

    rev_NICE_NAME = {v: k for k, v in NICE_NAME.items()}

    if culture_id in rev_NICE_NAME:
        culture_id = rev_NICE_NAME[culture_id]

    return votes, num_agents, params, culture_id



#!/usr/bin/env python

from typing import Union

import numpy as np

import mapel.roommates.cultures.euclidean as euclidean
import mapel.roommates.cultures.impartial as impartial
import mapel.roommates.cultures.mallows as mallows
import mapel.roommates.cultures.urn as urn


registered_roommates_culture = {
    'ic': impartial.generate_roommates_ic_votes,
    'id': impartial.generate_roommates_id_votes,
    'chaos': impartial.generate_roommates_chaos_votes,
    'symmetric': impartial.generate_roommates_symmetric_votes,
    'asymmetric': impartial.generate_roommates_asymmetric_votes,
    'urn': urn.generate_roommates_urn_votes,
    'fame': euclidean.generate_roommates_fame_votes,
    'expectation': euclidean.generate_roommates_expectation_votes,
    'attributes': euclidean.generate_roommates_attributes_votes,
    'euclidean': euclidean.generate_roommates_euclidean_votes,
    'reverse_euclidean': euclidean.generate_roommates_reverse_euclidean_votes,
    'group_ic': impartial.generate_roommates_group_ic_votes,
    'norm-mallows': mallows.generate_roommates_norm_mallows_votes,
    'mallows_euclidean': euclidean.generate_roommates_mallows_euclidean_votes,
    'malasym': mallows.generate_roommates_malasym_votes,
}


def generate_votes(culture_id: str = None, num_agents: int = None,
                   params: dict = None) -> Union[list, np.ndarray]:

    if culture_id in registered_roommates_culture:
        return registered_roommates_culture.get(culture_id)(num_agents=num_agents, **params)

    else:
        print("No such election culture_id!", culture_id)
        return []


# # # # # # # # # # # # # # # #
# LAST CLEANUP ON:  9.06.2023 #
# # # # # # # # # # # # # # # #






import numpy as np
from mapel.roommates.cultures._utils import convert
from mapel.core.utils import *
import logging


def generate_roommates_ic_votes(num_agents: int = None, **kwargs):
    """ Impartial Culture """

    votes = [list(np.random.permutation(num_agents)) for _ in range(num_agents)]

    return convert(votes)


def generate_roommates_group_ic_votes(num_agents: int = None,
                                      proportion=0.5,
                                      **kwargs):
    """ Impartial Culture with two groups """

    size_1 = int(proportion * num_agents)
    size_2 = int(num_agents - size_1)

    votes_1 = [list(np.random.permutation(size_1)) +
               list(np.random.permutation([j for j in range(size_1, num_agents)]))
               for _ in range(size_1)]

    votes_2 = [list(np.random.permutation([j for j in range(size_1, num_agents)])) +
               list(np.random.permutation(size_1))
               for _ in range(size_2)]

    votes = votes_1 + votes_2

    return convert(votes)


def generate_roommates_id_votes(num_agents: int = None, **kwargs):
    """ One of four extreme points for Compass """

    votes = [list(range(num_agents)) for _ in range(num_agents)]

    return convert(votes)


def generate_roommates_asymmetric_votes(num_agents: int = None, **kwargs):
    """ One of four extreme points for Compass """
    votes = [list(range(num_agents)) for _ in range(num_agents)]

    votes = [rotate(vote, shift) for shift, vote in enumerate(votes)]

    return convert(votes)


def generate_roommates_symmetric_votes(num_agents: int = None, **kwargs):
    """ One of four extreme points for Compass """

    num_rounds = num_agents - 1

    def next(agents):
        first = agents[0]
        last = agents[-1]
        middle = agents[1:-1]
        new_agents = [first, last]
        new_agents.extend(middle)
        return new_agents

    agents = [i for i in range(num_agents)]
    rounds = []

    for _ in range(num_rounds):
        pairs = []
        for i in range(num_agents // 2):
            agent_1 = agents[i]
            agent_2 = agents[num_agents - 1 - i]
            pairs.append([agent_1, agent_2])
        rounds.append(pairs)
        agents = next(agents)

    votes = np.zeros([num_agents, num_agents - 1], dtype=int)

    for pos, partition in enumerate(rounds):
        for x, y in partition:
            votes[x][pos] = y
            votes[y][pos] = x

    return votes


def generate_roommates_chaos_votes(num_agents: int = None, **kwargs):
    """ One of four extreme points for Compass """

    if num_agents-1 % 3 == 0:
        logging.warning("Incorrect realization of Chaos instance")

    num_rooms = num_agents // 2
    matrix = np.zeros([num_agents, num_agents - 1], dtype=int)

    matrix[0] = [i for i in range(num_agents - 1)]

    for i in range(1, num_agents):
        for j in range(num_rooms):
            matrix[i][2 * j] = (i + j - 1) % (num_agents - 1)
            if j < num_rooms - 1:
                matrix[i][2 * j + 1] = (num_rooms + i + j - 1) % (num_agents - 1)

    votes = np.zeros([num_agents, num_agents - 1], dtype=int)

    for k1 in range(num_agents):
        for k2 in range(num_agents - 1):
            for i in range(num_agents):
                if k1 != i and matrix[i][matrix[k1][k2]] == matrix[k1][k2]:
                    votes[k1][k2] = i

    return votes


# # # # # # # # # # # # # # # #
# LAST CLEANUP ON: 16.03.2022 #
# # # # # # # # # # # # # # # #




from .objects.RoommatesExperiment import RoommatesExperiment
from .objects.Roommates import Roommates
from . import cultures_ as rom
from .distances_ import get_distance


def prepare_online_roommates_experiment(**kwargs):
    return prepare_roommates_experiment(**kwargs, is_exported=False, is_imported=False)


def prepare_offline_roommates_experiment(**kwargs):
    return prepare_roommates_experiment(**kwargs, is_exported=True, is_imported=True)


def prepare_roommates_experiment(experiment_id=None,
                                 # instance_type=None,
                                 distance_id=None,
                                 is_imported=None,
                                 # is_shifted=False,
                                 is_exported=None,
                                 embedding_id=None):

    return RoommatesExperiment(experiment_id=experiment_id,
                               is_imported=is_imported,
                               distance_id=distance_id,
                               # instance_type=instance_type,
                               embedding_id=embedding_id,
                               # is_shifted=is_shifted,
                               is_exported=is_exported)


def generate_roommates_instance(**kwargs):
    instance = Roommates('virtual', 'tmp', **kwargs)
    instance.prepare_instance()
    return instance


def generate_roommates_votes(**kwargs):
    return rom.generate_votes(**kwargs)

def compute_distance(*args, **kwargs):
    return get_distance(*args, **kwargs)


#!/usr/bin/env python
import numpy as np
import copy
from mapel.core.objects.Instance import Instance

from mapel.roommates.cultures_ import generate_votes

from mapel.roommates.persistence.instance_imports import import_real_instance
from mapel.roommates.persistence.instance_exports import export_instance_to_a_file

from mapel.roommates.features_ import get_local_feature

class Roommates(Instance):

    def __init__(self,
                 experiment_id,
                 instance_id,
                 alpha=1,
                 culture_id=None,
                 num_agents=None,
                 is_imported=True,
                 votes=None):

        super().__init__(experiment_id, instance_id, alpha=alpha, culture_id=culture_id)

        self.num_agents = num_agents
        self.votes = votes

        self.retrospetive_vectors = None
        self.positionwise_vectors = None

        if is_imported and experiment_id != 'virtual':
            try:
                self.votes, self.num_agents, self.params, self.culture_id = \
                    import_real_instance(experiment_id)
                self.alpha = self.params['alpha']
            except:
                pass

    def get_retrospective_vectors(self):
        if self.retrospetive_vectors is not None:
            return self.retrospetive_vectors
        return self.votes_to_retrospective_vectors()

    def get_positionwise_vectors(self):
        if self.positionwise_vectors is not None:
            return self.positionwise_vectors
        return self.votes_to_positionwise_vectors()

    def votes_to_retrospective_vectors(self):

        vectors = np.zeros([self.num_agents, self.num_agents - 1], dtype=int)

        order_votes = [[] for _ in range(self.num_agents)]

        for a in range(self.num_agents):
            (missing,) = set(range(self.num_agents)) - set(self.votes[a])
            order_votes[missing] = copy.deepcopy(self.votes[a])

        for a in range(self.num_agents):
            for i, b in enumerate(order_votes[a]):
                vectors[a][i] = int(list(order_votes[b]).index(a))

        self.retrospetive_vectors = vectors
        return vectors

    def votes_to_positionwise_vectors(self):

        vectors = np.zeros([self.num_agents, self.num_agents - 1])

        for i in range(self.num_agents):
            pos = 0
            for j in range(self.num_agents - 1):
                vote = self.votes[i][j]
                vectors[vote][pos] += 1
                pos += 1

        for i in range(self.num_agents):
            for j in range(self.num_agents - 1):
                vectors[i][j] /= float(self.num_agents)

        self.positionwise_vectors = vectors
        return vectors

    def votes_to_pairwise_matrix(self) -> np.ndarray:
        """ convert VOTES to pairwise MATRIX """
        matrix = np.zeros([self.num_agents, self.num_agents])

        for v in range(self.num_agents):
            for c1 in range(self.num_agents - 1):
                for c2 in range(c1 + 1, self.num_agents - 1):
                    matrix[int(self.votes[v][c1])][int(self.votes[v][c2])] += 1

        for i in range(self.num_agents):
            for j in range(i + 1, self.num_agents):
                matrix[i][j] /= float(self.num_agents)
                matrix[j][i] = 1. - matrix[i][j]

        return matrix

    def prepare_instance(self, is_exported=None, params: dict = None):

        if params is None:
            params = {}

        if self.culture_id == 'roommates_norm-mallows' and 'norm-phi' not in params:
            params['norm-phi'] = np.random.rand()
            params['alpha'] = params['norm-phi']

        elif self.culture_id == 'roommates_urn' and 'alpha' not in params:
            params['alpha'] = np.random.rand()

        elif 'alpha' not in params:
            params['alpha'] = 1

        if 'variable' in params:
            params['alpha'] = params[params['variable']]

        self.alpha = params['alpha']

        self.votes = generate_votes(culture_id=self.culture_id,
                                    num_agents=self.num_agents,
                                    params=params)
        self.params = params

        if is_exported:
            export_instance_to_a_file(self)

    def compute_feature(self, feature_id, feature_long_id=None, **kwargs):
        if feature_long_id is None:
            feature_long_id = feature_id
        feature = get_local_feature(feature_id)
        self.features[feature_long_id] = feature(self, **kwargs)




import os
import pickle

import numpy as np
from mapel.roommates.cultures._utils import convert
from mapel.core.utils import *

import mapel.core.features.mallows as ml


def generate_mallows_votes(*args, **kwargs):
    return ml.generate_mallows_votes(*args, **kwargs)



def calculateZpoly(m):
    res = [1]
    for i in range(1, m + 1):
        mult = [1] * i
        res2 = [0] * (len(res) + len(mult) - 1)
        for o1, i1 in enumerate(res):
            for o2, i2 in enumerate(mult):
                res2[o1 + o2] += i1 * i2
        res = res2
    return res


def evaluatePolynomial(coeff, x):
    res = 0
    for i, c in enumerate(coeff):
        res += c * (x ** i)
    return res


def calculateZ(m, phi):
    coeff = calculateZpoly(m)
    return evaluatePolynomial(coeff, phi)


# mat[i][j] is the probability with which candidate i ends up in position j
def mallowsMatrix(num_candidates, lphi, pos, normalize=True):
    mat = np.zeros([num_candidates, num_candidates])
    if normalize:
        phi = ml.phi_from_normphi(num_candidates, lphi)
    else:
        phi = lphi
    Z = calculateZ(num_candidates, phi)
    for i in range(num_candidates):
        for j in range(num_candidates):
            freqs = [pos[k][i][j] for k in
                     range(1 + int(num_candidates * (num_candidates - 1) / 2))]
            unnormal_prob = evaluatePolynomial(freqs, phi)
            mat[i][j] = unnormal_prob / Z
    return mat


def get_mallows_matrix(num_candidates, params, normalize=True):
    lphi = params['norm-phi']
    weight = params['weight']
    # print(lphi, weight)
    try:
        path = os.path.join(os.getcwd(), 'mapel', 'voting', 'elections', 'mallows_positionmatrices',
                            str(num_candidates) + "_matrix.txt")
        print(path)
        with open(path, "rb") as file:
            pos = pickle.load(file)
    except FileNotFoundError:
        print("Mallows matrix only supported for up to 30 candidates")
    # print(pos)
    mat1 = mallowsMatrix(num_candidates, lphi, pos, normalize)
    res = np.zeros([num_candidates, num_candidates])
    for i in range(num_candidates):
        for j in range(num_candidates):
            res[i][j] = weight * mat1[i][j] + (1 - weight) * mat1[i][num_candidates - 1 - j]
    return res



def runif_in_simplex(n):
  ''' Return uniformly random vector in the n-simplex '''

  k = np.random.exponential(scale=1.0, size=n)
  return k / sum(k)


def generate_roommates_norm_mallows_votes(num_agents: int = None,
                                          normphi: float = 0.5,
                                          weight: float = 0,
                                          **kwargs):

    phi = ml.phi_from_normphi(num_agents, normphi=normphi)

    votes = generate_mallows_votes(num_agents, num_agents, phi=phi, weight=weight)

    return convert(votes)


def mallows_vote(vote, phi):
    num_candidates = len(vote)
    raw_vote = generate_mallows_votes(1, num_candidates, phi=phi, weight=0)[0]
    new_vote = [0] * len(vote)
    for i in range(num_candidates):
        new_vote[raw_vote[i]] = vote[i]
    return new_vote


def mallows_votes(votes, phi):
    for i in range(len(votes)):
        votes[i] = mallows_vote(votes[i], phi)
    return votes


def generate_roommates_malasym_votes(num_agents: int = None,
                                     normphi=0.5,
                                     **kwargs):
    """ Mallows on top of Asymmetric instance """

    votes = [list(range(num_agents)) for _ in range(num_agents)]

    votes = [rotate(vote, shift) for shift, vote in enumerate(votes)]

    # if 'norm-phi' not in params:
    #     params['norm-phi'] = np.random.rand()

    phi = ml.phi_from_normphi(num_agents, normphi=normphi)
    votes = mallows_votes(votes, phi)

    return convert(votes)


import numpy as np
import math
from numpy import linalg
from mapel.roommates.cultures._utils import convert
from mapel.roommates.cultures.mallows import mallows_votes

def get_range(params):
    if params['p_dist'] == 'beta':
        return np.random.beta(params['a'], params['b'])
    elif params['p_dist'] == 'uniform':
        return np.random.uniform(low=params['a'], high=params['b'])


def weighted_l1(a1, a2, w):
    total = 0
    for i in range(len(a1)):
        total += abs(a1[i] - a2[i]) * w[i]
    return total


def generate_roommates_attributes_votes(num_agents: int = None,
                                        dim: int = 2,
                                        space='uniform',
                                        **kwargs):
    name = f'{dim}d_{space}'

    agents_skills = np.array([get_rand(name) for _ in range(num_agents)])
    agents_weights = np.array([get_rand(name) for _ in range(num_agents)])

    votes = np.zeros([num_agents, num_agents], dtype=int)
    distances = np.zeros([num_agents, num_agents], dtype=float)
    ones = np.ones([dim], dtype=float)

    for v in range(num_agents):
        for c in range(num_agents):
            votes[v][c] = c
            if dim == 1:
                distances[v][c] = abs(1. - agents_skills[c]) * agents_weights[v]
            else:
                distances[v][c] = weighted_l1(ones, agents_skills[c], agents_weights[v])
        votes[v] = [x for _, x in sorted(zip(distances[v], votes[v]))]

    return convert(votes)


def generate_roommates_euclidean_votes(num_agents: int = None,
                                       dim: int = 2,
                                       space='uniform',
                                       **kwargs):
    name = f'{dim}d_{space}'

    agents = np.array([get_rand(name, i=i, num_agents=num_agents) for i in range(num_agents)])

    votes = np.zeros([num_agents, num_agents], dtype=int)
    distances = np.zeros([num_agents, num_agents], dtype=float)

    for v in range(num_agents):
        for c in range(num_agents):
            votes[v][c] = c
            distances[v][c] = np.linalg.norm(agents[v] - agents[c])
        votes[v] = [x for _, x in sorted(zip(distances[v], votes[v]))]

    return convert(votes)


def generate_roommates_reverse_euclidean_votes(num_agents: int = None,
                                               dim: int = 2,
                                               space='uniform',
                                               proportion=0.5,
                                               **kwargs):
    name = f'{dim}d_{space}'

    agents = np.array([get_rand(name, i=i, num_agents=num_agents) for i in range(num_agents)])

    votes = np.zeros([num_agents, num_agents], dtype=int)
    distances = np.zeros([num_agents, num_agents], dtype=float)

    for v in range(num_agents):
        for c in range(num_agents):
            votes[v][c] = c
            distances[v][c] = np.linalg.norm(agents[v] - agents[c])
        votes[v] = [x for _, x in sorted(zip(distances[v], votes[v]))]

    p = proportion

    for i in range(int(num_agents * (1. - p))):
        tmp = list(votes[i])
        tmp.reverse()
        votes[i] = tmp

    return convert(votes)


def generate_roommates_expectation_votes(num_agents: int = None,
                                         dim: int = 2,
                                         space='uniform',
                                         std=0.1,
                                         **kwargs):
    name = f'{dim}d_{space}'

    agents_reality = np.array([get_rand(name) for _ in range(num_agents)])
    agents_wishes = np.zeros([num_agents, 2])

    for v in range(num_agents):
        # while agents_wishes[v][0] <= 0 or agents_wishes[v][0] >= 1:
        agents_wishes[v][0] = np.random.normal(agents_reality[v][0], std)
        # while agents_wishes[v][1] <= 0 or agents_wishes[v][1] >= 1:
        agents_wishes[v][1] = np.random.normal(agents_reality[v][1], std)

    votes = np.zeros([num_agents, num_agents], dtype=int)
    distances = np.zeros([num_agents, num_agents], dtype=float)

    for v in range(num_agents):
        for c in range(num_agents):
            votes[v][c] = c
            distances[v][c] = np.linalg.norm(agents_reality[c] - agents_wishes[v])
        votes[v] = [x for _, x in sorted(zip(distances[v], votes[v]))]

    return convert(votes)


def generate_roommates_fame_votes(num_agents: int = None,
                                  dim: int = 2,
                                  space='uniform',
                                  radius=0.1,
                                  **kwargs):
    # Also known as radius model

    name = f'{dim}d_{space}'

    agents = np.array([get_rand(name) for _ in range(num_agents)])
    votes = np.zeros([num_agents, num_agents], dtype=int)
    distances = np.zeros([num_agents, num_agents], dtype=float)
    rays = np.array([np.random.uniform(0, radius) for _ in range(num_agents)])

    for v in range(num_agents):
        for c in range(num_agents):
            votes[v][c] = c
            distances[v][c] = np.linalg.norm(agents[v] - agents[c])
            distances[v][c] = distances[v][c] - rays[c]
        votes[v] = [x for _, x in sorted(zip(distances[v], votes[v]))]

    return convert(votes)


def generate_roommates_mallows_euclidean_votes(num_agents: int = None,
                                               dim: int = 2,
                                               space='uniform',
                                               phi=0.5,
                                               **kwargs):
    name = f'{dim}d_{space}'

    agents = np.array([get_rand(name, i=i, num_agents=num_agents) for i in range(num_agents)])

    votes = np.zeros([num_agents, num_agents], dtype=int)
    distances = np.zeros([num_agents, num_agents], dtype=float)

    for v in range(num_agents):
        for c in range(num_agents):
            votes[v][c] = c
            distances[v][c] = np.linalg.norm(agents[v] - agents[c])
        votes[v] = [x for _, x in sorted(zip(distances[v], votes[v]))]

    votes = mallows_votes(votes, phi)

    return convert(votes)

# AUXILIARY
def random_ball(dimension, num_points=1, radius=1):
    random_directions = np.random.normal(size=(dimension, num_points))
    random_directions /= linalg.norm(random_directions, axis=0)
    random_radii = np.random.random(num_points) ** (1 / dimension)
    return radius * (random_directions * random_radii).T

GEN_CTR = 0
def get_rand(model: str, i: int = 0, num_agents: int = 0, cat: str = "voters") -> list:
    """ generate random values"""
    # print(model ==  "1d_uniform")

    point = [0]
    if model in {"1d_uniform", "1d_interval"}:
        return np.random.rand()
    elif model in {'1d_asymmetric'}:
        if np.random.rand() < 0.3:
            return np.random.normal(loc=0.25, scale=0.15, size=1)
        else:
            return np.random.normal(loc=0.75, scale=0.15, size=1)
    elif model in {"1d_gaussian"}:
        point = np.random.normal(0.5, 0.15)
        while point > 1 or point < 0:
            point = np.random.normal(0.5, 0.15)
    elif model == "1d_one_sided_triangle":
        point = np.random.uniform(0, 1) ** 0.5
    elif model == "1d_full_triangle":
        point = np.random.choice(
            [np.random.uniform(0, 1) ** 0.5, 2 - np.random.uniform(0, 1) ** 0.5])
    elif model == "1d_two_party":
        point = np.random.choice([np.random.uniform(0, 1), np.random.uniform(2, 3)])
    elif model in {"2d_disc", "2d_range_disc"}:
        phi = 2.0 * 180.0 * np.random.random()
        radius = math.sqrt(np.random.random()) * 0.5
        point = [0.5 + radius * math.cos(phi), 0.5 + radius * math.sin(phi)]
    elif model == "2d_range_overlapping":
        phi = 2.0 * 180.0 * np.random.random()
        radius = math.sqrt(np.random.random()) * 0.5
        if cat == "voters":
            point = [0.25 + radius * math.cos(phi), 0.5 + radius * math.sin(phi)]
        elif cat == "candidates":
            point = [0.75 + radius * math.cos(phi), 0.5 + radius * math.sin(phi)]
    elif model in {"2d_square", "2d_uniform"}:
        point = [np.random.random(), np.random.random()]
    elif model in {'2d_asymmetric'}:
        if np.random.rand() < 0.3:
            return np.random.normal(loc=0.25, scale=0.15, size=2)
        else:
            return np.random.normal(loc=0.75, scale=0.15, size=2)
    elif model == "2d_sphere":
        alpha = 2 * math.pi * np.random.random()
        x = 1. * math.cos(alpha)
        y = 1. * math.sin(alpha)
        point = [x, y]
    elif model in ["2d_gaussian", "2d_range_gaussian"]:
        point = [np.random.normal(0.5, 0.15), np.random.normal(0.5, 0.15)]
        while np.linalg.norm(point - np.array([0.5, 0.5])) > 0.5:
            point = [np.random.normal(0.5, 0.15), np.random.normal(0.5, 0.15)]
    elif model in ["2d_range_fourgau"]:
        r = np.random.randint(1, 4)
        size = 0.06
        if r == 1:
            point = [np.random.normal(0.25, size), np.random.normal(0.5, size)]
        if r == 2:
            point = [np.random.normal(0.5, size), np.random.normal(0.75, size)]
        if r == 3:
            point = [np.random.normal(0.75, size), np.random.normal(0.5, size)]
        if r == 4:
            point = [np.random.normal(0.5, size), np.random.normal(0.25, size)]
    elif model in ["3d_cube", "3d_uniform"]:
        point = [np.random.random(), np.random.random(), np.random.random()]
    elif model in ["5d_uniform"]:
        dim = 5
        point = [np.random.random() for _ in range(dim)]
    elif model in ["10d_uniform"]:
        dim = 10
        point = [np.random.random() for _ in range(dim)]
    elif model in {'3d_asymmetric'}:
        if np.random.rand() < 0.3:
            return np.random.normal(loc=0.25, scale=0.15, size=3)
        else:
            return np.random.normal(loc=0.75, scale=0.15, size=3)
    elif model in ['3d_gaussian']:
        point = [np.random.normal(0.5, 0.15),
                 np.random.normal(0.5, 0.15),
                 np.random.normal(0.5, 0.15)]
        while np.linalg.norm(point - np.array([0.5, 0.5, 0.5])) > 0.5:
            point = [np.random.normal(0.5, 0.15),
                     np.random.normal(0.5, 0.15),
                     np.random.normal(0.5, 0.15)]
    elif model == "4d_cube":
        dim = 4
        point = [np.random.random() for _ in range(dim)]
    elif model == "5d_cube":
        dim = 5
        point = [np.random.random() for _ in range(dim)]
    elif model == '1d_extreme':
        if i % 2 == 1:
            i -= 0.1
        point = i
    elif model == '2d_extreme':
        if i % 2 == 1:
            alpha = 2 * math.pi * ((i - np.random.random() / 100) / num_agents)
        else:
            alpha = 2 * math.pi * ((i + np.random.random() / 100) / num_agents)
        x = 1. * math.cos(alpha)
        y = 1. * math.sin(alpha)
        point = [x, y]
    else:
        print('unknown culture_id', model)
        point = [0, 0]
    return point



#!/usr/bin/env python

import queue
from itertools import chain

import numpy as np
from scipy.special import binom
from mapel.roommates.cultures._utils import convert, remove_first


#### new ####
def next_sig(sig):
    new_sig = [0] * len(sig)
    overflow = 1
    for i in range(0, len(sig)):
        if sig[i] == 0 and overflow == 0:
            new_sig[i] = 0
        elif sig[i] == 0 and overflow == 1:
            new_sig[i] = 1
            overflow = 0
        elif sig[i] == 1 and overflow == 0:
            new_sig[i] = 1
        elif sig[i] == 1 and overflow == 1:
            new_sig[i] = 0
            overflow = 1
    return new_sig

import math
def generate_roommates_gs_ideal_votes(num_agents=None, params=None):

    m = num_agents
    n = num_agents

    decomposition_tree = _balanced(m)

    all_inner_nodes = get_all_inner_nodes(decomposition_tree)

    SIG = np.zeros([n, len(all_inner_nodes)])
    for s in range(1, n):
        SIG[s] = next_sig(SIG[s-1])

    EXT_SIG = np.zeros([n, len(all_inner_nodes)])
    for i, sig in enumerate(SIG):
        ctr = 0
        for k in range(int(math.log(n, 2))):
            # print('k', k, int(n/2**(k+1)))
            for t in range(int(n/2**(k+1))):
                # print('t', t)
                EXT_SIG[i][ctr] = sig[k]
                ctr += 1
                # EXT_SIG[k][2*i] = sig[i]
                # EXT_SIG[k][2*i+1] = sig[i]

    # print(SIG)
    # print(EXT_SIG)

    compute_depth(decomposition_tree)

    votes = []
    for i in range(n):

        signature = list(SIG[i])
        # signature  = [0,0,1,0,0,0,0]
        # signature.reverse()
        # print(signature)

        # print(SIG[i])

        for j, node in enumerate(all_inner_nodes):
            node.reverse = SIG[i][node.depth]
            # print(node.reverse)

        raw_vote = sample_a_vote(decomposition_tree)
        vote = [int(candidate.replace('x', '')) for candidate in raw_vote]
        votes.append(vote)

        for i, node in enumerate(all_inner_nodes):
            node.reverse = False

    # print(votes)
    # print(remove_first(votes))

    # order votes
    votes = sorted(votes, key=lambda x: x[0])

    return convert(votes)


def generate_roommates_revgs_ideal_votes(num_agents=None, params=None):

    m = num_agents
    n = num_agents

    decomposition_tree = _balanced(m)

    all_inner_nodes = get_all_inner_nodes(decomposition_tree)

    SIG = np.zeros([n, len(all_inner_nodes)])
    for s in range(1, n):
        SIG[s] = next_sig(SIG[s-1])

    EXT_SIG = np.zeros([n, len(all_inner_nodes)])
    for i, sig in enumerate(SIG):
        ctr = 0
        for k in range(int(math.log(n, 2))):
            # print('k', k, int(n/2**(k+1)))
            for t in range(int(n/2**(k+1))):
                # print('t', t)
                EXT_SIG[i][ctr] = sig[k]
                ctr += 1
                # EXT_SIG[k][2*i] = sig[i]
                # EXT_SIG[k][2*i+1] = sig[i]

    # print(SIG)
    # print(EXT_SIG)

    compute_depth(decomposition_tree)

    votes = []
    for i in range(n):

        signature = list(SIG[i])
        # signature  = [0,0,1,0,0,0,0]
        # signature.reverse()
        # print(signature)

        print(SIG[i])

        for j, node in enumerate(all_inner_nodes):
            node.reverse = SIG[i][node.depth]
            # print(node.reverse)

        raw_vote = sample_a_vote(decomposition_tree)
        vote = [int(candidate.replace('x', '')) for candidate in raw_vote]
        votes.append(vote)

        for i, node in enumerate(all_inner_nodes):
            node.reverse = False

    # print(votes)
    # print(remove_first(votes))

    votes = remove_first(votes)

    for i in range(0, num_agents, 2):
        votes[i+1].reverse()

    votes = [rotate(vote, shift) for shift, vote in enumerate(votes)]

    return votes

# HELPER
def rotate(vector, shift):
    return vector[shift:] + vector[:shift]


def generate_roommates_gs_votes(num_agents=None, params=None):

    if params is None:
        params = {}

    if params is not None and 'tree' not in params:
        params = {'tree': 'random'}

    while True:
        m = num_agents
        n = num_agents

        if params['tree'] == 'random':
            func = lambda m, r: 1./(m-1) * binom(m - 1, r) * \
                                binom(m - 1 + r, m)
            buckets = [func(m, r) for r in range(1, m)]

            denominator = sum(buckets)
            # print(buckets)
            buckets = [buckets[i]/denominator for i in range(len(buckets))]

            num_internal_nodes = \
                np.random.choice(len(buckets), 1, p=buckets)[0]+1

            decomposition_tree = \
                _decompose_tree(num_agents, num_internal_nodes)

        elif params['tree'] == 'caterpillar':
            decomposition_tree = _caterpillar(m)

        elif params['tree'] == 'balanced':
            decomposition_tree = _balanced(m)

        all_inner_nodes = get_all_inner_nodes(decomposition_tree)

        votes = []
        for i in range(n):

            signature = [np.random.choice([0, 1])
                         for _ in range(len(all_inner_nodes))]

            for i, node in enumerate(all_inner_nodes):
                node.reverse = signature[i]

            raw_vote = sample_a_vote(decomposition_tree)
            vote = [int(candidate.replace('x', '')) for candidate in raw_vote]
            votes.append(vote)

            for i, node in enumerate(all_inner_nodes):
                node.reverse = False

        # return votes, decomposition_tree
        return convert(votes)


#### copy from 'elections' project ####
def _decompose_tree(num_leaves, num_internal_nodes):
    """ Algorithm from: Uniform generation of a Schroder tree"""

    num_nodes = num_leaves + num_internal_nodes

    patterns = _generate_patterns(num_nodes, num_internal_nodes)
    seq, sizes = _generate_tree(num_nodes, num_internal_nodes, patterns)

    seq = cycle_lemma(seq)

    tree = _turn_pattern_into_tree(seq)

    return tree


def generate_ordinal_group_separable_votes(num_voters=None, num_candidates=None, params=None):
    """ Algorithm from: The Complexity of Election Problems
    with Group-Separable Preferences"""

    if params is None:
        params = {}

    if params is not None and 'tree' not in params:
        params = {'tree': 'random'}

    while True:
        m = num_candidates
        n = num_voters

        if params['tree'] == 'random':
            func = lambda m, r: 1./(m-1) * binom(m - 1, r) * \
                                binom(m - 1 + r, m)
            buckets = [func(m, r) for r in range(1, m)]

            denominator = sum(buckets)
            # print(buckets)
            buckets = [buckets[i]/denominator for i in range(len(buckets))]

            num_internal_nodes = \
                np.random.choice(len(buckets), 1, p=buckets)[0]+1

            decomposition_tree = \
                _decompose_tree(num_candidates, num_internal_nodes)

        elif params['tree'] == 'caterpillar':
            decomposition_tree = _caterpillar(m)

        elif params['tree'] == 'balanced':
            decomposition_tree = _balanced(m)

        all_inner_nodes = get_all_inner_nodes(decomposition_tree)

        votes = []
        for _ in range(n):

            signature = [np.random.choice([0, 1])
                         for _ in range(len(all_inner_nodes))]

            for i, node in enumerate(all_inner_nodes):
                node.reverse = signature[i]

            raw_vote = sample_a_vote(decomposition_tree)
            vote = [int(candidate.replace('x', '')) for candidate in raw_vote]
            votes.append(vote)

            for i, node in enumerate(all_inner_nodes):
                node.reverse = False

        # return votes, decomposition_tree
        return votes

REVERSE = {}

def compute_depth(root):
    root.depth = 0
    q = queue.Queue()
    q.put(root)
    while not q.empty():
        node = q.get()
        for child in node.children:
            child.depth = node.depth+1
            q.put(child)


def get_all_leaves_names(node):
    if node.leaf:
        return [node.election_id]
    output = []
    for i in range(len(node.children)):
        output.append(get_all_leaves_names(node.children[i]))
    return list(chain.from_iterable(output))


def get_all_leaves_nodes(node):
    if node.leaf:
        return [node]
    output = []
    for i in range(len(node.children)):
        output.append(get_all_leaves_nodes(node.children[i]))
    return list(chain.from_iterable(output))


def get_all_nodes(root):
    all_nodes = []
    q = queue.Queue()
    q.put(root)
    while not q.empty():
        node = q.get()
        all_nodes.append(node)
        for child in node.children:
            q.put(child)
    return all_nodes


def get_bracket_notation(node):
    if node.leaf:
        return str(node.election_id)
    output = ''
    for i in range(len(node.children)):
        output += str(get_bracket_notation(node.children[i]))
    return '(' + output + ')'


def get_all_inner_nodes(node):
    if node.leaf:
        return []
    output = [[node]]
    for i in range(len(node.children)):
        output.append(get_all_inner_nodes(node.children[i]))
    return list(chain.from_iterable(output))


def sample_a_vote(node, reverse=False):
    if node.leaf:
        return [node.election_id]
    output = []
    if reverse == node.reverse:
        for i in range(len(node.children)):
            output.append(sample_a_vote(node.children[i]))
    else:
        for i in range(len(node.children)):
            output.append(sample_a_vote(node.children[len(node.children)-1-i],
                                        reverse=True))

    return list(chain.from_iterable(output))


class Node:

    total_num_leaf_descendants = 0

    def __init__(self, election_id):

        self.election_id = election_id
        self.parent = None
        self.children = []
        self.leaf = True
        self.reverse = False

        self.left = 0
        self.right = 0

        self.num_leaf_descendants = None
        self.depth = None
        self.scheme = {}
        self.scheme_1 = {}
        self.scheme_2 = {}
        self.vector = []

    def __str__(self):
        return self.election_id

    def add_child(self, child):
        child.parent = self
        self.children.append(child)
        self.leaf = False


def _generate_patterns(num_nodes, num_internal_nodes):
    # Step 1: Mixing the patterns
    patterns = ['M0' for _ in range(num_nodes-num_internal_nodes)] + \
               ['M1' for _ in range(num_internal_nodes)]
    np.random.shuffle(patterns)
    return patterns


def _generate_tree(num_nodes, num_internal_nodes, patterns):
    """ Algorithm from: A linear-time embedding_id for the generation of trees """

    sequence = []
    sizes = []
    larges = []
    ctr = 0
    inner_ctr = 0
    for i, pattern in enumerate(patterns):
        if pattern == 'M0':
            sequence.append('x' + str(ctr))
            sizes.append(1)
            ctr += 1
        elif pattern == 'M1':
            sequence.append('v' + str(inner_ctr))
            sequence.append('()1')   # instead of 'o'
            sequence.append('f1')
            sequence.append('f1')
            sizes.append(4)
            larges.append(i)
            inner_ctr += 1

    num_classical_edges = 0
    num_semi_edges = 2*num_internal_nodes
    num_multi_edges = num_internal_nodes
    num_trees = 1
    pos = 1
    num_edges = num_nodes - num_trees - num_semi_edges - num_classical_edges

    pos_to_insert = []
    for i, elem in enumerate(sequence):
        if elem == '()1':
            pos_to_insert.append(i)

    choices = list(np.random.choice([i for i in range(len(pos_to_insert))],
                                    size=num_edges, replace=True))
    choices.sort(reverse=True)

    for choice in choices:
        sizes[larges[choice]] += 1
        sequence.insert(pos_to_insert[choice], 'f1')

    for i in range(len(pos_to_insert)):
        sequence.remove('()1')

    return sequence, sizes


def _turn_pattern_into_tree(pattern):
    stack = []
    for i, element in enumerate(pattern):
        if 'x' in element or 'v' in element:
            stack.append(Node(element))
        if 'f' in element:
            parent = stack.pop()
            child = stack.pop()
            parent.add_child(child)
            stack.append(parent)
    return stack[0]


def cycle_lemma(sequence):

    pos = 0
    height = 0
    min = 0
    pos_min = 0
    for element in sequence:
        if 'x' in element or 'v' in element:
            if height <= min:
                pos_min = pos
                min = height
            height += 1
        if 'f' in element:
            height -= 1
        pos += 1

    # rotate
    for _ in range(pos_min):
        element = sequence.pop(0)
        sequence.append(element)

    return sequence


def _add_num_leaf_descendants(node):
    """ add total number of descendants to each internal node """

    if node.leaf:
        node.num_leaf_descendants = 1
    else:
        node.num_leaf_descendants = 0
        for child in node.children:
            node.num_leaf_descendants += _add_num_leaf_descendants(child)

    return node.num_leaf_descendants


def _add_scheme(node):

    # print(node.election_id)
    for starting_pos in node.scheme_1:

        pos = starting_pos
        for child in node.children:
            if pos in child.scheme_1:
                child.scheme_1[pos] += node.scheme_1[starting_pos]
            else:
                child.scheme_1[pos] = node.scheme_1[starting_pos]
            pos += child.num_leaf_descendants

    for starting_pos in node.scheme_2:
        # print(starting_pos)
        pos = starting_pos
        for child in node.children:
            if pos in child.scheme_2:
                child.scheme_2[pos] += node.scheme_2[starting_pos]
            else:
                child.scheme_2[pos] = node.scheme_2[starting_pos]
            pos -= child.num_leaf_descendants

    if node.leaf:
        _construct_vector_from_scheme(node)
    else:
        for child in node.children:
            _add_scheme(child)


def _construct_vector_from_scheme(node):

    x = node.scheme_1
    y = node.scheme_2
    node.scheme = {k: x.get(k, 0) + y.get(k, 0) for k in set(x) | set(y)}
    # print(node.scheme, x, y)

    weight = 1. / sum(node.scheme.values())

    node.vector = [0 for _ in range(Node.total_num_leaf_descendants)]
    for key in node.scheme:

        node.vector[int(key)] += node.scheme[key] * weight


# def get_frequency_matrix_from_tree(root):
#
#     _add_num_leaf_descendants(root)
#
#     Node.total_num_leaf_descendants = root.num_leaf_descendants
#     root.scheme_1 = {0: 1}
#     root.scheme_2 = {int(Node.total_num_leaf_descendants - 1): 1}
#
#     _add_scheme(root)
#
#     nodes = get_all_leaves_nodes(root)
#
#     m = Node.total_num_leaf_descendants
#     array = np.zeros([m, m])
#
#     for i in range(m):
#         for j in range(m):
#             array[i][j] = nodes[i].vector[j]
#
#     return array


def _caterpillar(num_leaves):
    root = Node('root')
    tmp_root = root
    ctr = 0

    while num_leaves > 2:
        leaf = Node('x' + str(ctr))
        inner_node = Node('v' + str(ctr))
        tmp_root.add_child(leaf)
        tmp_root.add_child(inner_node)
        tmp_root = inner_node
        num_leaves -= 1
        ctr += 1

    leaf_1 = Node('x' + str(ctr))
    leaf_2 = Node('x' + str(ctr + 1))
    tmp_root.add_child(leaf_1)
    tmp_root.add_child(leaf_2)

    return root


def _balanced(num_leaves):
    root = Node('root')
    ctr = 0

    q = queue.Queue()
    q.put(root)

    # while ctr < num_leaves-2:
    while q.qsize() * 2 < num_leaves:
        tmp_root = q.get()
        for _ in range(2):
            inner_node = Node('v' + str(ctr))
            tmp_root.add_child(inner_node)
            q.put(inner_node)
            ctr += 1

    ctr = 0
    while ctr < num_leaves:
        tmp_root = q.get()
        for _ in range(2):
            node = Node('x' + str(ctr))
            tmp_root.add_child(node)
            ctr += 1

    # print_tree(root)
    return root


def print_tree(root):
    print(root.election_id)
    for child in root.children:
        print_tree(child)


### MATRICES ###
def get_gs_caterpillar_matrix(num_candidates):
    return get_gs_caterpillar_vectors(num_candidates).transpose()


def get_gs_caterpillar_vectors(num_candidates):
    return get_frequency_matrix_from_tree(_caterpillar(num_candidates))

# get_gs_caterpillar_matrix(10)

# 10votes = generate_group_separable_election(num_voters=10, num_candidates=14)
# print(votes)

# generate_group_separable_election(num_voters=100,num_candidates=40)


def set_left_and_right(node):
    for i, child in enumerate(node.children):
        left = 0
        right = 0
        for j, other_child in enumerate(node.children):
            if j < i:
                left += other_child.num_leaf_descendants
            elif j > i:
                right += other_child.num_leaf_descendants
        child.left = left
        child.right = right

    for child in node.children:
        set_left_and_right(child)


def get_frequency_matrix_from_tree(root):
    _add_num_leaf_descendants(root)

    m = int(root.num_leaf_descendants)

    f = {}

    all_nodes = get_all_nodes(root)
    for node in all_nodes:
        print(node.election_id)
        f[str(node.election_id)] = [0 for _ in range(m)]
    set_left_and_right(root)

    f[root.election_id][0] = 1

    for node in all_nodes:
        if node.election_id != root.election_id:
            # print(node.election_id)
            for t in range(m):
                value_1 = 0
                if t-node.left >= 0:
                    value_1 = 0.5*f[node.parent.election_id][t - node.left]
                value_2 = 0
                if t-node.right >= 0:
                    value_2 = 0.5*f[node.parent.election_id][t - node.right]
                f[str(node.election_id)][t] = value_1 + value_2

    vectors = []
    for i in range(m):
        name = 'x' + str(i)
        vectors.append(f[name])
    return np.array(vectors)


# num_voters = 1000
# num_candidates = 5
# votes, tree = generate_group_separable_election(num_voters=num_voters,
#                                       num_candidates=num_candidates)
#
# print(get_frequency_matrix_from_tree(tree))
#
# vectors = np.zeros([num_candidates, num_candidates])
#
# for i in range(num_voters):
#     pos = 0
#     for j in range(num_candidates):
#         vote = votes[i][j]
#         if vote == -1:
#             continue
#         vectors[vote][pos] += 1
#         pos += 1
# for i in range(num_candidates):
#     for j in range(num_candidates):
#         vectors[i][j] /= float(num_voters)
#
# print(vectors)



#!/usr/bin/env python

import mapel.roommates.features.basic_features as basic
from mapel.core.glossary import MAIN_GLOBAL_FEATUERS
from mapel.core.features_main import get_main_global_feature
import numpy as np
from itertools import combinations
from mapel.core.inner_distances import l2

registered_roommates_features = {
    'summed_rank_minimal_matching': basic.summed_rank_minimal_matching,
    'summed_rank_maximal_matching': basic.summed_rank_maximal_matching,
    'min_num_bps_matching': basic.min_num_bps_matching,
    'num_of_bps_min_weight': basic.num_of_bps_maximumWeight,
    'avg_num_of_bps_for_rand_matching': basic.avg_num_of_bps_for_random_matching,
    'mutuality': basic.mutuality,
    'dist_from_id_1': basic.dist_from_id_1,
    'dist_from_id_2': basic.dist_from_id_2,
    'minimal_rank_maximizing_matching': basic.minimal_rank_maximizing_matching,
}


def get_local_feature(feature_id):
    if feature_id in registered_roommates_features:
        return registered_roommates_features.get(feature_id)
    return {
    }.get(feature_id)


def get_global_feature(feature_id):
    if feature_id in MAIN_GLOBAL_FEATUERS:
        return get_main_global_feature(feature_id)

    return {'monotonicity': monotonicity,
            'distortion_from_all': distortion_from_all,
            }.get(feature_id)

# TMP FUNCS
def monotonicity(experiment, instance) -> float:
    e0 = instance.instance_id
    c0 = np.array(experiment.coordinates[e0])
    distortion = 0
    for e1, e2 in combinations(experiment.instances, 2):
        if e1 != e0 and e2 != e0:
            original_d1 = experiment.distances[e0][e1]
            original_d2 = experiment.distances[e0][e2]
            original_proportion = original_d1 / original_d2
            embedded_d1 = np.linalg.norm(c0 - experiment.coordinates[e1])
            embedded_d2 = np.linalg.norm(c0 - experiment.coordinates[e2])
            embedded_proportion = embedded_d1 / embedded_d2
            _max = max(original_proportion, embedded_proportion)
            _min = min(original_proportion, embedded_proportion)
            distortion += _max / _min
    return distortion


def distortion_from_all(experiment, election):
    values = np.array([])
    one_side_values = np.array([])
    election_id_1 = election.instance_id

    for election_id_2 in experiment.instances:
        if election_id_1 != election_id_2:
            true_distance = experiment.distances[election_id_1][election_id_2]
            true_distance /= experiment.distances['MD']['MA']
            embedded_distance = l2(np.array(experiment.coordinates[election_id_1]),
                                   np.array(experiment.coordinates[election_id_2]))

            embedded_distance /= \
                l2(np.array(experiment.coordinates['MD']),
                   np.array(experiment.coordinates['MA']))
            one_side_ratio = embedded_distance / true_distance
            one_side_values = np.append(one_side_values, one_side_ratio)

            ratio = max(embedded_distance, true_distance) / min(embedded_distance, true_distance)

            values = np.append(values, ratio)

    return np.mean(values)




try:
    import gurobipy as gp
    from gurobipy import GRB
except:
    pass

try:
    from mapel.roommates.matching.games import StableRoommates
except:
    pass

from random import shuffle
import statistics
import warnings
import sys
import time
import networkx as nx

sys.setrecursionlimit(10000)
# warnings.filterwarnings("error")

def generate_instance(num_agents):
    instance=[]
    for i in range(num_agents):
        pref = [x for x in range(num_agents) if x != i]
        shuffle(pref)
        instance.append(pref)
    return instance

def number_blockingPairs(instance,matching):
    bps=0
    num_agents=len(instance)
    for i in range(num_agents):
        for j in range(i+1,num_agents):
            if i!=j:
                partner_i=matching[i]
                partneri_index=instance[i].index(partner_i)
                partner_j = matching[j]
                partnerj_index = instance[j].index(partner_j)
                if instance[i].index(j)<partneri_index:
                    if instance[j].index(i) < partnerj_index:
                        bps+=1
    return bps


def compute_stable_SR(votes):
    dict_instance={}
    num_agents=len(votes)
    for i in range(num_agents):
        dict_instance[i]=votes[i]
    game = StableRoommates.create_from_dictionary(dict_instance)
    try:
        matching = game.solve()
        usable_matching = {}
        for m in matching:
            usable_matching[m.name] = matching[m].name
        return usable_matching
    except:
        return 'None'



def spear_distance(instance1,instance2):
    num_agents=len(instance1)
    m = gp.Model("qp")
    #m.setParam('Threads', 6)
    #m.setParam('OutputFlag', False)
    x = m.addVars(num_agents, num_agents, lb=0, ub=1, vtype=GRB.BINARY)
    opt = m.addVar(vtype=GRB.INTEGER, lb=0, ub=num_agents * num_agents*num_agents)
    for i in range(num_agents):
        m.addConstr(gp.quicksum(x[i, j] for j in range(num_agents)) == 1)
        m.addConstr(gp.quicksum(x[j, i] for j in range(num_agents)) == 1)
    m.addConstr(gp.quicksum(abs(instance1[i].index(k)-instance2[j].index(t)) * x[i, j]*x[k,t] for i in range(num_agents)
                            for j in range(num_agents) for k in [x for x in range(num_agents) if x != i]
                            for t in [x for x in range(num_agents) if x != j]) == opt)
    m.setObjective(opt, GRB.MINIMIZE)
    m.optimize()
    for i in range(num_agents):
        for j in range(num_agents):
            if x[i, j].X == 1:
                print(i,j)
    return int(m.objVal)


def spear_distance_linear(instance1,instance2):
    num_agents=len(instance1)
    m = gp.Model("mip1")
    #m.setParam('Threads', 6)
    #m.setParam('OutputFlag', False)
    x = m.addVars(num_agents, num_agents, lb=0, ub=1, vtype=GRB.BINARY)
    y = m.addVars(num_agents, num_agents,num_agents,num_agents, lb=0, ub=1, vtype=GRB.CONTINUOUS)
    opt = m.addVar(vtype=GRB.INTEGER, lb=0, ub=num_agents * num_agents*num_agents)
    for i in range(num_agents):
        m.addConstr(gp.quicksum(x[i, j] for j in range(num_agents)) == 1)
        m.addConstr(gp.quicksum(x[j, i] for j in range(num_agents)) == 1)
    for i in range(num_agents):
        for j in range(num_agents):
            for k in range(num_agents):
                for l in range(num_agents):
                    m.addConstr(y[i,j,k,l]<=x[i, j])
                    m.addConstr(y[i, j, k, l] <= x[k, l])
                    m.addConstr(y[i, j, k, l] >= x[i, j]+x[k, l]-1)

    m.addConstr(gp.quicksum(abs(instance1[i].index(k)-instance2[j].index(t)) * y[i, j,k,t] for i in range(num_agents)
                            for j in range(num_agents) for k in [x for x in range(num_agents) if x != i]
                            for t in [x for x in range(num_agents) if x != j]) == opt)
    m.setObjective(opt, GRB.MINIMIZE)
    m.optimize()
    for i in range(num_agents):
        for j in range(num_agents):
            if x[i, j].X == 1:
                print(i,j)
    return int(m.objVal)

#Only call for instances that admit a stable matchig
#summed: Set to true we try to optimize the summed rank of agents for their partner in the matching, set to false we optimize the minimum rank
#Best (only relevant if summed is set to true): Set to true we output the best possible matching, set to false the worst one
def rank_matching(instance,best,summed):
    num_agents=len(instance)
    m = gp.Model("mip1")
    m.setParam('OutputFlag', False)
    m.setParam('Threads', 10)
    x = m.addVars(num_agents, num_agents, lb=0, ub=1, vtype=GRB.BINARY)
    opt = m.addVar(vtype=GRB.INTEGER, lb=0, ub=num_agents*num_agents)
    for i in range(num_agents):
        m.addConstr(x[i, i]  == 0)
    for i in range(num_agents):
        for j in range(num_agents):
            m.addConstr(x[i, j]  == x[j, i])
    for i in range(num_agents):
        m.addConstr(gp.quicksum(x[i, j] for j in range(num_agents)) == 1)
    for i in range(num_agents):
        for j in range(i+1,num_agents):
            better_pairs=[]
            for t in range(0,instance[i].index(j)+1):
                better_pairs.append([i,instance[i][t]])
            for t in range(0,instance[j].index(i)+1):
                better_pairs.append([j,instance[j][t]])
            m.addConstr(gp.quicksum(x[a[0], a[1]] for a in better_pairs) >= 1)
    if summed:
        #print(instance[0])
        m.addConstr(gp.quicksum(instance[i].index(j)* x[i, j]  for i in range(num_agents) for j in [x for x in range(num_agents) if x != i]) == opt)
        if best:
            m.setObjective(opt, GRB.MAXIMIZE)
        else:
            m.setObjective(opt, GRB.MINIMIZE)
    else:
        for i in range(num_agents):
            m.addConstr(gp.quicksum(instance[j].index(i) * x[i, j] for j in [x for x in range(num_agents) if x != i])<=opt)
        m.setObjective(opt, GRB.MINIMIZE)
    m.optimize()
    matching={}
    for i in range(num_agents):
        for j in range(num_agents):
            if abs(x[i, j].X - 1) < 0.05:
                matching[i]=j
                matching[j]=i
    return int(m.objVal), matching

def min_num_bps_matching(instance):
    num_agents=len(instance.votes)
    m = gp.Model("mip1")
    m.setParam('OutputFlag', False)
    x = m.addVars(num_agents, num_agents, lb=0, ub=1, vtype=GRB.BINARY)
    y = m.addVars(num_agents, num_agents, lb=0, ub=1, vtype=GRB.BINARY)
    opt = m.addVar(vtype=GRB.INTEGER, lb=0, ub=num_agents*num_agents)
    m.addConstr(gp.quicksum(y[i, j] for i in range(num_agents) for j in range(num_agents)) <= opt)
    for i in range(num_agents):
        m.addConstr(x[i, i]  == 0)
    for i in range(num_agents):
        for j in range(num_agents):
            m.addConstr(x[i, j]  == x[j, i])
    for i in range(num_agents):
        m.addConstr(gp.quicksum(x[i, j] for j in range(num_agents)) <= 1)
    for i in range(num_agents):
        for j in range(i+1,num_agents):
            better_pairs=[]
            for t in range(0,instance.votes[i].index(j)+1):
                better_pairs.append([i,instance.votes[i][t]])
            for t in range(0,instance.votes[j].index(i)+1):
                better_pairs.append([j,instance.votes[j][t]])
            m.addConstr(gp.quicksum(x[a[0], a[1]] for a in better_pairs) >= 1-y[i,j])

    m.setObjective(opt, GRB.MINIMIZE)
    m.optimize()
    matching={}
    for i in range(num_agents):
        for j in range(num_agents):
            if abs(x[i, j].X - 1) < 0.05:
                matching[i]=j
                matching[j]=i
    #return int(m.objVal), matching
    return int(m.objVal)

def summed_rank_maximal_matching(instance):
    val,matching= rank_matching(instance.votes,True,True)
    if number_blockingPairs(instance.votes,matching)>0:
        print('ERROR IN summed_rank_maximal_matching')
        exit(0)
    return val


def summed_rank_minimal_matching(instance):
    val,matching= rank_matching(instance.votes,False,True)
    if number_blockingPairs(instance.votes,matching)>0:
        print('ERROR IN summed_rank_minimal_matching')
        exit(0)
    return val

def minimal_rank_maximizing_matching(instance):
    val,matching=rank_matching(instance.votes,True,False)
    if number_blockingPairs(instance.votes,matching)>0:
        print('ERROR IN minimal_rank_maximizing_matching')
        exit(0)
    return val

def avg_num_of_bps_for_random_matching(instance,iterations=100):
    bps=[]
    num_agents=len(instance.votes)
    for _ in range(iterations):
        #create random matching
        agents=list(range(num_agents))
        shuffle(agents)
        matching=[agents[i:i + 2] for i in range(0, num_agents, 2)]
        matching_dict={}
        for m in matching:
            matching_dict[m[0]]=m[1]
            matching_dict[m[1]] = m[0]
        bps.append(number_blockingPairs(instance.votes,matching_dict))
    return statistics.mean(bps), statistics.stdev(bps)


def num_of_bps_maximumWeight(instance) -> int:
    num_agents=len(instance.votes)
    G = nx.Graph()
    for i in range(num_agents):
        for j in range(i+1,num_agents):
            G.add_edge(i,j,weight=2*(num_agents-1)-instance.votes[i].index(j)-instance.votes[j].index(i))
            #G.add_edge(i, j, weight=instance[i].index(j) + instance[j].index(i))
    matching = nx.max_weight_matching(G, maxcardinality=True)
    matching_dict = {}
    for p in matching:
        matching_dict[p[0]] = p[1]
        matching_dict[p[1]] = p[0]
    return number_blockingPairs(instance.votes,matching_dict)

# NEW 15.03.2022
def mutuality(instance) -> int:
    vectors = instance.get_retrospective_vectors()
    score = 0
    for vector in vectors:
        for i, v in enumerate(vector):
            score += abs(v-i)
    return score

def dist_from_id_1(instance) -> int:
    vectors = instance.get_retrospective_vectors()
    score = 0
    for vector in vectors:
        for i in range(len(vector)):
            for j in range(i,len(vector)):
                score += abs(vector[j]-vector[i])
    return score

def dist_from_id_2(instance) -> int:
    vectors = instance.get_retrospective_vectors()
    score = 0
    for vector in vectors:
        score += len(set(vector))
    return score